"""This module contains the state model of the chp."""

from typing import cast

from pysimmods.base.state import ModelState
from pysimmods.base.types import ModelInitVals


class CHPCNGState(ModelState):
    """A CHP CNG state information class.

    Parameters
    ----------
    inits: dict
        Contains the initial configuration of this CHP model. See
        attributes section for specific information.

    Attributes
    ----------
    active_s: int
        Time since unit switch to current power level in [s].
    active_s_per_day: int
        Time the unit is active on the current day in [s].
    inactive_s: int
        Time since the unit is switched off in [s].
    inactive_s_per_day: int
        Time the is inactive on the current day in [s].
    restarts: int
        Number of restarts on the current day.
    p_th_kw: float
        Thermal power generated in [kW].
    gas_cons_m3: float
        Gas consumed in the last step in [m^3]
    """

    def __init__(self, inits: ModelInitVals):
        super().__init__(inits)

        self.active_s: int = cast("int", inits["active_s"])
        self.active_s_per_day: int = cast("int", inits["active_s_per_day"])
        self.inactive_s: int = cast("int", inits["inactive_s"])
        self.inactive_s_per_day: int = cast("int", inits["inactive_s_per_day"])

        self.restarts: int = cast("int", inits["restarts"])

        self.p_th_kw: float = 0.0
        self.gas_cons_m3: float = 0.0



"""This module contains the config model for the PV plant."""

from typing import cast

from pysimmods.base.config import ModelConfig
from pysimmods.base.types import ModelParams


class PVConfig(ModelConfig):
    """Config parameters of PV plant model.

    Parameters
    ----------
    params : dict
        Contains the configuration of the PV plant. See attribute
        section for more information about the parameters, attributes
        marked with *(Input)* can or must be provided.

    Attributes
    ----------
    a_m2 : float
        (Input) The size of the surface of the PV plant in [m²].
    lat_deg : float, optional
        (Input) The geographical latitude of the plant position in [°].
        Default is the latitude of Oldenburg.
    lon_deg : float, optional
        (Input) The geographical longitude of the plant position in
        [°]. Default is the longitude of Oldenburg.
    orient_deg : float, optional
        (Input) The orientation of the plant in [°]. 0° means
        north and 90° means east. Default is 180° (south).
    tilt_deg : float, optional
        (Input) The tilt of the plant in [°]. Default is 30°.
    eta_percent : float, optional
        (Input) The efficiency of the PV plant at reference conditions
        (25°C module temperature, 1000 Watt/m² radiation power)
        including inverter losses, in [%]. Default is 25%.
    beta_percent_per_kelvin : float, optional
        (Input) The efficiency losses caused by module temperature
        exceeding the reference temperature of 25°C, in [%/K]. Default
        is 0.29.
    k_m_w_per_m2 : float, optional
        (Input) The half-saturation constant to the Michaelis-Menten
        saturation curve which is used to model the irradiance
        dependency of the PV plant efficiency in [W/m²]. Default is 50.
    alpha_w_per_m2_kelvin : float, optional
        (Input) The heat transmission coefficient between the PV plant
        and the air in [W/(m²K)]. Heat exchange is assumed to take only
        place on the top side of the PV plant. Default is 11.
    rho_kg_per_m2 : float, optional
        (Input) The mass per unit area of the PV plant in [kg/m²].
        Default is 15.
    reflex_percent : float, optional
        (Input) The reflexiveness of the surface area of the PV plant
        in [%]. Reflexiveness is assumed to be independent from angle
        of incidence. Default is 5.
    c_j_per_kw_kelvin : float, optional
        (Input) The specific heat capacity of the PV plant in [J/kg K].
        Default is 900.
    is_static_t_module : bool, optional
        If set to true the module temperature is kept constant
        during simulation at the temperature the initial temperature
        (see :class:`~.PVState`). Default is False.
    has_external_irradiance_model : bool, optional
        If set to True, an external irradiance model is assumed and
        instead for :attr:`~.PVInputs.bh_w_per_m2` and
        :attr:`~.PVInputs.dh_w_per_m2` input for
        :attr:`~.PVInputs.s_module_w_per_m2` has to be provided,
        default is False.
    p_peak_kw : float
        Peak power of the PV plant in [kW], measured under standard
        testing conditions (module temperature 25°C, irradiance 1000 W)

    """

    def __init__(self, params: ModelParams):
        super().__init__(params)
        self.a_m2: float = cast("float", params["a_m2"])

        self.lat_deg: float = cast("float", params.get("lat_deg", 53.143890))
        self.lon_deg: float = cast("float", params.get("lon_deg", 8.213889))
        self.orient_deg: float = cast("float", params.get("orient_deg", 180))
        self.tilt_deg: float = cast("float", params.get("tilt_deg", 30))
        self.eta_percent: float = cast("float", params.get("eta_percent", 25))
        self.beta_percent_per_kelvin: float = cast(
            "float", params.get("beta_percent_per_kelvin", 0.29)
        )
        self.k_m_w_per_m2: float = cast(
            "float", params.get("k_m_w_per_m2", 50)
        )
        self.alpha_w_per_m2_kelvin: float = cast(
            "float", params.get("alpha_w_per_m2_kelvin", 11)
        )
        self.rho_kg_per_m2: float = cast(
            "float", params.get("rho_kg_per_m2", 15)
        )
        self.reflex_percent: float = cast(
            "float", params.get("reflex_percent", 5)
        )
        self.c_j_per_kg_kelvin: float = cast(
            "float", params.get("c_j_per_kg_kelvin", 900)
        )
        self.is_static_t_module: bool = cast(
            "bool", params.get("is_static_t_module", False)
        )
        self.has_external_irradiance_model: bool = cast(
            "bool", params.get("has_external_irradiance_model", False)
        )

        self.p_max_kw: float = self.eta_percent / 100 * self.a_m2
        self.p_min_kw: float = 0
        self.max_step_size: int = 300
        self.default_p_schedule = None
        self.default_q_schedule = None



from typing import Generic

from pysimmods.base.model import C, I, Model, S


class Buffer(Model[C, S, I], Generic[C, S, I]):
    """A buffer subtype model.

    This class provides all required functions for a buffer so that
    derived models only need to provide the step-function. However,
    those functions can be overwritten if needed.

    """

    def get_pn_charge_max_kw(self) -> float:
        """Return nominal maximum charging rate in kW.

        Respects the sign convention with charging rate as positive
        values.
        """
        if hasattr(self.config, "p_charge_max_kw"):
            return getattr(self.config, "p_charge_max_kw") * self.config.lsign
        else:
            return self.config.p_max_kw * self.config.lsign

    def get_pn_charge_min_kw(self) -> float:
        """Return nominal minimum charging rate in kW.

        Respects the sign convention with charging rate as positive
        values.
        """
        if hasattr(self.config, "p_charge_min_kw"):
            return getattr(self.config, "p_charge_min_kw") * self.config.lsign
        else:
            return self.config.p_min_kw * self.config.lsign

    def get_pn_discharge_max_kw(self) -> float:
        """Return nominal maximum discharging rate in kW.

        Respects the sign convention with discharging rate as negative
        values.
        """
        if hasattr(self.config, "p_discharge_max_kw"):
            return (
                getattr(self.config, "p_discharge_max_kw") * self.config.gsign
            )
        else:
            return self.config.p_max_kw * self.config.gsign

    def get_pn_discharge_min_kw(self) -> float:
        """Return nominal minimum charging rate in kW.

        Respects the sign convention with discharging rate as negative
        values.
        """

        if hasattr(self.config, "p_discharge_min_kw"):
            return (
                getattr(self.config, "p_discharge_min_kw") * self.config.gsign
            )
        else:
            return self.config.p_min_kw * self.config.gsign

    def get_pn_min_kw(self) -> float:
        """Minimum power of battery in kW.

        The minimum power is the maximum discharging power in passive
        sign convention.

        """
        if self.config.psc:
            return self.get_pn_discharge_max_kw()
        else:
            return self.get_pn_charge_max_kw()

    def get_pn_max_kw(self) -> float:
        """Maximum power of battery in kW.

        The maximum power is the maximum charging power in passive sign
        convention.

        """
        if self.config.psc:
            return self.get_pn_charge_max_kw()
        else:
            return self.get_pn_discharge_max_kw()

    def set_p_kw(self, p_kw: float) -> None:
        """Set the target active power value.

        With passive sign convention, positive values indicate charging
        and *p_kw* needs to be between pn_charge_min_kw and
        pn_charge_max_kw. Negative values indicate discharging and
        *p_kw* needs to be between pn_discharge_min_kw and
        pn_discharge_max_kw.

        If *p_kw* is lower than the respective min value, it will be set
        to zero. If it is higher than the respective max value, it will
        be set to the max value.

        """
        if self.config.psc and p_kw < 0 or self.config.asc and p_kw > 0:
            # Discharging
            if p_kw < 0:
                p_max = self.get_pn_discharge_min_kw()
                p_min = self.get_pn_discharge_max_kw()
            else:
                p_min = self.get_pn_discharge_min_kw()
                p_max = self.get_pn_discharge_max_kw()
        elif self.config.psc and p_kw > 0 or self.config.asc and p_kw < 0:
            # Charging
            if p_kw < 0:
                p_max = self.get_pn_charge_min_kw()
                p_min = self.get_pn_charge_max_kw()
            else:
                p_min = self.get_pn_charge_min_kw()
                p_max = self.get_pn_charge_max_kw()
        else:
            self.inputs.p_set_kw = 0
            return

        if abs(p_kw) < abs(p_min) and abs(p_kw) < abs(p_max):
            self.inputs.p_set_kw = 0
        else:
            self.inputs.p_set_kw = max(p_min, min(p_max, p_kw))

    def set_q_kvar(self, q_kvar) -> None:
        """Set the target reactive power value.

        Currently, there is no further functionality behind this,
        because there is no buffer model that makes use of reactive
        power.

        """
        self.inputs.q_set_kvar = q_kvar

    def get_p_kw(self) -> float:
        """Return the current active power in kW of this model.

        Respects the current sign convention.
        """
        return self.state.p_kw * self.config.lsign



"""
This module contains a reimplementation of the MATLAB battery model
provided by the TU Munich in context of the research project iHEM
(intelligent Home Energy Management).

"""

import math
from copy import copy

from pysimmods.base.buffer import Buffer
from pysimmods.base.types import ModelInitVals, ModelParams
from pysimmods.buffer.batterysim.config import BatteryConfig
from pysimmods.buffer.batterysim.inputs import BatteryInputs
from pysimmods.buffer.batterysim.state import BatteryState


class Battery(Buffer[BatteryConfig, BatteryState, BatteryInputs]):
    """Simple battery simulation model

    Self-discharge and aging are not considered. Effect of charging power on
    efficiency eta is modelled by fitting a polynomial model to data measured
    by TU Munich.

    See::

        Schimpe, M.; Piesch, C.; Hesse, H.C.; Paß, J.; Ritter, S.; Jossen, A.
        Power Flow Distribution Strategy for Improved Power Electronics
        Energy Efficiency in Battery Storage Systems: Development and
        Implementation in a Utility-Scale System. Energies 2018, 11, 533.
        https://doi.org/10.3390/en11030533

    You have to provide the two dictionaries *params* and *inits*.
    *params* provides the configuration parameters for the battery model and
    might look like this::

        {
            "cap_kwh": 5,
            "p_charge_max_kw": 1,
            "p_discharge_max_kw": 1,
            "soc_min_percent": 15,
            "eta_pc": [-2.109566, 0.403556, 97.110770],
        }

    Here *cap_kwh* is the electric capacity of the battery in kWh. *p_min_kw*
    and *p_max_kw* specify the minimum and maximum power of the battery in kW.
    Negative values indicate charging (battery is a consumer).  Positive values
    indicate discharging (battery is a producer). *soc_min_percent* indicates
    the minimum state of charge in percent below which discharging is stopped.
    The entry *eta_pc* is a list and contains the coefficients a, b, c of a
    quadratic polynomial function, which is used to model power dependency of
    efficiency.

    The dict *inits* provides initial values for state variables. The
    battery model has only one state variable that must be specified when the
    model is initialized. It is *soc_percent*, which indicates the initial
    state of charge in percent of battery capacity. The model has two more
    state variables *p_kw* and *eta_percent*. They indicate the current power
    and efficiency of the battery. But as they are flow quantities and their
    current value has no effect in the next simulation step, they are just set
    to None during initialization of the model::

        {"soc_percent": 50}

    Attributes
    ----------
    config : :class:`~.BatteryConfig`
        Stores the configuration parameters of the battery model.
    state : :class:`~.BatteryState`
        Stores the initialization parameters of the battery model.
    inputs : :class:`~.BatteryInputs`
        Stores the input parameters for each step of the battery
        model.

    """

    def __init__(self, params: ModelParams, inits: ModelInitVals):
        self.config = BatteryConfig(params)
        self.state = BatteryState(inits)
        self.inputs = BatteryInputs()

    def step(self, pretend: bool = False) -> BatteryState:
        """Perform a simulation step."""

        nstate = copy(self.state)
        self._check_inputs(nstate)

        self._calculate_efficiency(nstate)

        if nstate.p_kw * self.config.gsign > 0:
            self._discharge(nstate)
        else:
            self._charge(nstate)

        nstate.soc_percent = nstate._energy_kwh / self.config.cap_kwh * 100

        if not pretend:
            self.state = nstate
        return nstate

    def _check_inputs(self, nstate: BatteryState):
        """Check constraints for active power."""
        if self.inputs.p_set_kw is None:
            if self.inputs.now_dt is not None:
                nstate.p_kw = self.config.default_p_schedule[
                    self.inputs.now_dt.hour
                ]
            else:
                nstate.p_kw = 0
        else:
            # discharge
            if self.inputs.p_set_kw * self.config.gsign > 0:
                if self.config.p_discharge_max_kw > abs(self.inputs.p_set_kw):
                    nstate.p_kw = self.inputs.p_set_kw
                else:
                    nstate.p_kw = self.config.p_discharge_max_kw
            # charge
            else:
                if self.config.p_charge_max_kw > abs(self.inputs.p_set_kw):
                    nstate.p_kw = self.inputs.p_set_kw
                else:
                    nstate.p_kw = self.config.p_charge_max_kw
            # copy sign to allow all sign conventions
            nstate.p_kw = math.copysign(nstate.p_kw, self.inputs.p_set_kw)

    def _calculate_efficiency(self, nstate: BatteryState):
        """Calculate efficiency using a second degree polynomial."""

        nstate._energy_kwh = self.config.cap_kwh * self.state.soc_percent / 100
        p_set_norm = nstate.p_kw / self.config.cap_kwh

        nstate.eta_percent = (
            self.config.eta_pc[0] * p_set_norm**2
            + self.config.eta_pc[1] * p_set_norm
            + self.config.eta_pc[2]
        )

    def _discharge(self, nstate: BatteryState):
        delta_energy_kwh = (
            nstate.p_kw
            / (nstate.eta_percent / 100)
            * (self.inputs.step_size / 3600)
        )

        theoretical_energy_kwh = (
            nstate._energy_kwh
            - self.config.cap_kwh * self.config.soc_min_percent / 100
        )

        if theoretical_energy_kwh > abs(delta_energy_kwh):
            # Won't be fully discharged in this step
            nstate._energy_kwh += delta_energy_kwh
        else:
            # Will be fully discharged in this step
            nstate.p_kw = theoretical_energy_kwh / (
                self.inputs.step_size / 3600 / nstate.eta_percent * -100
            )
            nstate._energy_kwh = (
                self.config.soc_min_percent / 100 * self.config.cap_kwh
            )

    def _charge(self, nstate: BatteryState):
        delta_energy_kwh = (
            nstate.p_kw
            * (nstate.eta_percent / 100)
            * (self.inputs.step_size / 3600)
        )

        if (self.config.cap_kwh - nstate._energy_kwh) > delta_energy_kwh:
            # Won't be fully charged in this step
            nstate._energy_kwh += delta_energy_kwh
        else:
            # Will be fully charged in this step
            nstate.p_kw = (self.config.cap_kwh - nstate._energy_kwh) / (
                self.inputs.step_size / 3600 * nstate.eta_percent / 100
            )
            nstate._energy_kwh = self.config.cap_kwh



from enum import Enum

from typing_extensions import TypeAlias, Union


class QControl(Enum):
    PRIORITIZE_P = 0
    PRIORITIZE_Q = 1


ModelParams: TypeAlias = dict[
    str, Union[float, int, str, bool, list[float] | QControl, "ModelParams"]
]

ModelInitVals: TypeAlias = dict[
    str, Union[float, int, str, bool, list[float], "ModelInitVals"]
]



"""This module contains the input information of the heat demand
model.
"""

from pysimmods.base.model import ModelInputs


class HeatDemandInputs(ModelInputs):
    """Captures the inputs of the heat demand model"""

    def __init__(self):
        super().__init__()

        self.day_avg_t_air_deg_celsius: float = 0.0



from pysimmods.other.heatdemandsim.heatdemand import HeatDemand as HeatDemand
from pysimmods.other.heatdemandsim.presets import (
    heatdemand_preset as heatdemand_preset,
)
from pysimmods.other.invertersim.inverter import Inverter as Inverter



from typing import Generic

from pysimmods.base.model import C, I, Model, S


class Consumer(Model[C, S, I], Generic[C, S, I]):
    """A consumer subtype model.

    A consumer returns positive power values in the
    consumer reference arrow system (passive sign convention).

    """

    def set_p_kw(self, p_kw: float | None) -> None:
        """Set the target value for active power in kW.

        Since the model internally ignores the sign convention,
        the value is converted to a positive value.

        If None is provided, the model will fallback to the default
        schedule.
        """
        if p_kw is not None:
            self.inputs.p_set_kw = abs(p_kw)
        else:
            self.inputs.p_set_kw = None

    def get_p_kw(self) -> float:
        """Return the current active power in kW of the model.

        Respects the current sign convention.
        """
        return self.state.p_kw * self.config.lsign



"""This module contains a :class:`mosaik_api.Simulator` for all models
of the pysimmods package.

"""

import json
import logging
import warnings
from datetime import datetime, timedelta, timezone
from typing import Any, Dict

import mosaik_api_v3
import numpy as np
from midas.util.logging import set_and_init_logger
from mosaik_api_v3.types import (
    CreateResult,
    InputData,
    Meta,
    ModelName,
    OutputData,
    OutputRequest,
    SimId,
    Time,
)
from typing_extensions import override

from pysimmods.base.model import Model
from pysimmods.mosaik_bridge import LOG
from pysimmods.mosaik_bridge.meta import META, MODELS
from pysimmods.util.date_util import GER

VLOG = logging.getLogger("pysimmods.mosaik_bridge.verbose")


class PysimmodsSimulator(mosaik_api_v3.Simulator):
    """The Pysimmods simulator."""

    def __init__(self):
        super().__init__(META)
        self.sid: str = ""
        self.step_size: int = 0
        self.now_dt: datetime = datetime(1977, 1, 1)
        self.key_value_logs: bool

        self.models: dict[str, Model] = {}
        self.num_models: dict[str, int] = {}

    @override
    def init(
        self,
        sid: SimId,
        time_resolution: float = 1.0,
        start_date: str = "2010-01-01 00:00:00+0100",
        step_size: int = 900,
        **sim_params,  # for compatibility
    ) -> Meta:
        """Called exactly ones after the simulator has been started.

        Parameters
        ----------
        sid : str
            Simulator ID for this simulator.
        start_date : str
            The start date as UTC ISO 8601 date string.
        step_size : int, optional
            Step size for this simulator. Defaults to 900.
        key_value_logs: bool, optional
            If set to True, a different, json-based logging format will
            be used. Default is False.
        time_resolution: float, optional
            Something new from mosaik 3.

        Returns
        -------
        dict
            The meta dict (set by *mosaik_api.Simulator*).

        """
        self.sid = sid
        self.step_size = step_size
        self.now_dt = datetime.strptime(start_date, GER).astimezone(
            timezone.utc
        )

        return self.meta

    @override
    def create(
        self,
        num: int,
        model: ModelName,
        *,
        params: dict[str, Any],
        inits: dict[str, Any],
        **model_params,  # for compatibility
    ) -> list[CreateResult]:
        """Initialize the simulation model instance (entity).

        Parameters
        ----------
        num: int
            The number of models to create.
        model: str
            The name of the models to create. Must be present inside
            the simulator's meta.
        params: Dict[str, Any]
            The dictionary with parameters for the model to be created.
            Must be provided but can be empty.
        inits: Dict[str, Any]
            The dictionary with initial values for the state of the
            model to be created. Must be provided but can be empty.

        Returns
        -------
        list
            A list with information on the created entity.

        """
        entities = []
        self.num_models.setdefault(model, 0)

        for _ in range(num):
            eid = f"{model}-{self.num_models[model]}"
            self.models[eid] = MODELS[model](params, inits)
            self.num_models[model] += 1
            entities.append({"eid": eid, "type": model})

        return entities

    @override
    def step(
        self, time: Time, inputs: InputData, max_advance: Time = 0
    ) -> Time | None:
        """Perform a simulation step.

        Parameters
        ----------
        time : int
            The current simulation step (by convention in seconds since
            simulation start.
        inputs : dict
            A *dict* containing inputs for entities of this simulator.
        max_advance: int,
            Required for mosaik>=3 but not explicitly supported yet.

        Returns
        -------
        int
            The next step this simulator wants to be stepped.

        """
        self._set_default_inputs()

        # Set inputs from other simulators
        for eid, attrs in inputs.items():
            for attr, src_ids in attrs.items():
                self._log_input(eid, attr, src_ids)

                # Use time information from time generator
                if attr == "local_time":
                    self._set_attr_local_time(eid, src_ids)
                    continue

                attr_sum = self._aggregate_attr(src_ids)
                self._set_remaining_attrs(eid, attr, attr_sum)

        # Step the models
        for model in self.models.values():
            model.step()

        # Update time for the next step
        self.now_dt += timedelta(seconds=self.step_size)

        return time + self.step_size

    @override
    def get_data(self, outputs: OutputRequest) -> OutputData:
        """Return the requested output (if feasible).

        Parameters
        ----------
        outputs : dict
            A *dict* containing requested outputs of each entity.

        Returns
        -------
        dict
            A *dict* containing the values of the requested outputs.

        """

        data = {}
        for eid, attrs in outputs.items():
            for attr in attrs:
                value = self._get_remaining_attrs(eid, attr)
                data.setdefault(eid, dict())[attr] = value

                self._log_output(eid, attr, value)

        return data

    def _set_default_inputs(self):
        VLOG.debug(
            "Setting step size %d and current time %s to all models.",
            self.step_size,
            self.now_dt,
        )
        for _, model in self.models.items():
            model.set_step_size(self.step_size)
            model.set_now_dt(self.now_dt)

    def _log_input(self, eid: str, attr: str, src_ids: Dict[str, Any]):
        for src_id, val in src_ids.items():
            if isinstance(val, np.ndarray):
                val = val.tolist()
            elif isinstance(val, np.integer):
                val = int(val)
            elif isinstance(val, np.floating):
                val = float(val)

            LOG.debug(
                json.dumps(
                    {
                        "id": f"{self.sid}_{eid}",
                        "name": eid,
                        "model": eid.split("-")[0],
                        "type": "input",
                        "attribute": attr,
                        "source": src_id,
                        "value": val,
                    }
                )
            )

    def _log_output(self, eid: str, attr: str, value: Any):
        LOG.debug(
            json.dumps(
                {
                    "id": f"{self.sid}_{eid}",
                    "name": eid,
                    "model": eid.split("-")[0],
                    "type": "output",
                    "attribute": attr,
                    "value": value,
                }
            )
        )

    def _set_attr_local_time(self, eid: str, src_ids: Dict[str, Any]) -> bool:
        for val in src_ids.values():
            self.models[eid].set_now_dt(val)
            self.now_dt = datetime.strptime(val, GER).astimezone(timezone.utc)
            return True

        return False

    def _aggregate_attr(self, src_ids: Dict[str, Any]) -> float:
        """Aggregate inputs from different sources.

        If more inputs for one source exists, the average is calculated.

        """
        attr_sum = 0
        for val in src_ids.values():
            if val is None:
                continue
            if isinstance(val, (list, np.ndarray)):
                # This should only happen if palaestrAI is used
                val = val[0]
            attr_sum += val
        attr_sum /= len(src_ids)

        return float(attr_sum)

    def _set_remaining_attrs(self, eid: str, attr: str, attr_sum: float):
        # Apply corrections
        if attr in ("p_set_mw", "p_th_set_mw", "q_set_mvar"):
            attr = attr.replace("m", "k")
            attr_sum *= 1e3

        # Set the inputs
        if attr == "set_percent":
            warnings.warn(
                "Using set_percent is deprecated and will be ignored. Use "
                "p_set_kw or p_set_mw instead",
                UserWarning,
            )

        elif attr == "p_set_kw":
            self.models[eid].set_p_kw(attr_sum)
        elif attr == "q_set_kvar":
            self.models[eid].set_q_kvar(attr_sum)
        else:
            setattr(self.models[eid].inputs, attr, attr_sum)

    def _get_remaining_attrs(self, eid: str, attr: str) -> float:
        # Apply correction of the attr if necessary
        if attr in ("p_mw", "p_th_mw", "q_mvar"):
            true_attr = attr.replace("m", "k")
        elif attr == "p_possible_max_mw":
            true_attr = "p_possible_max_kw"
        else:
            true_attr = attr

        if true_attr == "p_kw":
            value = self.models[eid].get_p_kw()
        elif true_attr == "q_kvar":
            value = self.models[eid].get_q_kvar()
        else:
            value = getattr(self.models[eid].state, true_attr)

        # Apply correction of the value if necessary
        if attr in ("p_mw", "p_th_mw", "q_mvar", "p_possible_max_mw"):
            value *= 1e-3

        return value


if __name__ == "__main__":
    set_and_init_logger(
        0, "pysimmods-logfile", "pysimmods-mosaik.log", replace=True
    )
    LOG.info("Starting mosaik simulation...")
    mosaik_api_v3.start_simulation(PysimmodsSimulator())



"""Contains the preset function for thermal household profiles

:constant CHP_AVERAGE_RUNTIME:
    source: 'IHK Projekte Hannover, Blockheizkraftwerke,
    Seite 5 IHK Projekte Durschnittliche Betriebsstunden
    eines BHKWs'
:constant CHP_MIN_TH_CONS_COEF:
    source: 'IHK Projekte Hannover, Blockheizkraftwerke,
    Seite 5 Prozentzahl für die untere minimale thermische
    Leistung des Jahresverbrauchs eines Householdtypes'
:constant CHP_MAX_TH_CONS_COEF:
    source: 'IHK Projekte Hannover, Blockheizkraftwerke,
    Seite 5 Prozentzahl für die obere maximale thermische
    Leistung des Jahresverbrauchs eines Householdtypes'
:constant DEGREE_OF_EFFIENCY:
    source: 'Niedertemperatur- und Brennwertkessel.
    Wissenswertes über moderne Zentralheizungsanlagen.
    Hessisches Ministerium für Umwelt, Energie,
    Landwirtschaft und Verbraucherschutz. 2011'

"""

from types import FunctionType
from typing import Any, cast

import numpy as np

from pysimmods.base.types import ModelInitVals, ModelParams
from pysimmods.other.heatdemandsim.util import hprofiles

CHP_AVERAGE_RUNTIME = 5_000.0
CHP_MIN_TH_CONS_COEF = 1.0
CHP_MAX_TH_CONS_COEF = 1.5
# DEGREE_OF_EFFIENCY = 0.93


def heatdemand_preset(
    pn_th_kw: float, seed: int | None = None, **kwargs: dict[str, Any]
) -> tuple[ModelParams, ModelInitVals]:
    rng = np.random.default_rng(seed)

    profile = find_demand_profile(rng, pn_th_kw)

    # Adapt the consumer constant of the household, so that power
    # and consumption better fit to each other
    profile["consumer_constant"] = pn_th_kw * 24 * 2 / 3
    inits: ModelInitVals = {
        "t_last_3_deg_celsius": 0.0,
        "t_last_2_deg_celsius": 0.0,
        "t_last_1_deg_celsius": 0.0,
    }
    return profile, inits


def demand_profiles() -> dict[str, ModelParams]:
    """Load heat demand profiles into global store."""

    profiles = {}
    for name, func in hprofiles.__dict__.items():
        if "__" not in name and isinstance(func, FunctionType):
            profiles[name] = func()

    return profiles


def find_demand_profile(
    rng: np.random.Generator, chp_p_th_kw: float
) -> ModelParams:
    """Find heat demand profile based on thermal power supply."""
    min_consumption = {}
    max_consumption = {}
    profiles = demand_profiles()
    for key, val in profiles.items():
        # average annual consumption per average operating
        # hours of the chp
        consumption = (
            cast("float", val["consumer_constant"]) * 365 / CHP_AVERAGE_RUNTIME
        )
        # calculate lower power boundary
        min_consumption[key] = consumption * CHP_MIN_TH_CONS_COEF
        # calculate upper power boundary
        max_consumption[key] = consumption * CHP_MAX_TH_CONS_COEF

    possible_profiles = []
    distances = {}

    # Look for fitting households
    for key, val in profiles.items():
        if abs(chp_p_th_kw >= min_consumption[key]) and abs(
            chp_p_th_kw <= max_consumption[key]
        ):
            # Profile fits optimally
            possible_profiles.append(key)
            distances[key] = 0.0
        else:
            # Profile does not fit, calculate difference
            lower = abs(abs(chp_p_th_kw) - min_consumption[key])
            upper = abs(abs(chp_p_th_kw) - max_consumption[key])
            distances[key] = lower if lower < upper else upper

    if len(possible_profiles) == 1:
        # Exactly one matching profile
        demand_profile = profiles[possible_profiles[0]].copy()
    elif len(possible_profiles) > 1:
        # More than one matching profile, select randomly
        demand_profile = profiles[rng.choice(possible_profiles)].copy()
    else:
        # No matching profile, select the nearest one

        min_type = "one_family"
        min_val = float("Inf")

        for key, val in distances.items():
            if val < min_val:
                min_val = val
                min_type = key

        demand_profile = profiles[min_type].copy()
    return demand_profile



"""This module provides configurations for the HVAC model.

The base formula is derived from the practical training energy
informatics at the University of Oldenburg. There is a frigde model
with 80 Watts and an air conditioning with 2 Kilowatts. To use one of
those, you can do it like this::

    params, inits = hvac_preset(0.08)  # for the fridge
    params, inits = hvac_preset(2)  # for the air conditioning

Additional, configurations for "TK-Anlagen" were derived from
https://www.tis-gdv.de/tis/tagungen/svt/svt10/weilhart/inhalt-htm/
The modeling of those should be considered as "educated guess" rather
than an acurate modeling. Those are available with different nominal
power values::

    hvac_preset(60)
    hvac_preset(100)
    hvac_preset(232)
    hvac_preset(415)
    hvac_preset(1743)
    hvac_preset(2075)
    hvac_preset(2325)
    hvac_preset(3139)
    hvac_preset(4456)

Choosing any of those will give a solid configuration but it is also
possible to choose any value between; the preset function will
determine the closest configuration to the provided nominal power.
Selecting a value that is located in the middle of two of those values
can lead to a model that has either not enough power or too much power,
resulting in a behavior that is further away from reality.

To overcome this, it is possible create configurations procedurally, by
setting either `randomize_params=True` and/or `randomize_inits=True`.
This will aim to create a configuration that is, even under worst case
conditions, able to provide the necessary power. However there is no
guarantee that cooling house with such a configuration could exist in
reality.

As an example how the calculations for the configuration is done, we
will go through the 60 kW model. All of the following calculations use
rounded values. According to the link above, the facility has an
energetic value of approx. 205 kWh/(m^3*a) and a storage volume of
800 m^3. This results in a total of 164000 kWh/a and 450 kWh/Day.

Now we're assuming that those energy is based on a workload of 33 %,
which means this cooling house could consume up to 1350 kWh/Day and,
consequently, 56.6 kWh/h. This is rounded up to 60 kW. However, since
the nominal power is the provided value, we need to reverse those
calculations to find a suitable configuration for a given power value.

The initialization parameters are derived from the air conditioning
configuration, which assumes 500 kg of oak furniture with a specific
heat capacity of 2390 J/(kg*K). This goes into the equation as
500 kg * 2390 J/(kg*K) = 1195000 J/K.

However, the cooling house does not store oak, but food. Non-frozen
food, e.g., apples, have a specific heat capacity of 3890 J(kg*K).
To get a similar cooling-power relationship, we calculate a
reference mass of 1195000 J/K / 3890 J/(kg*K) = 311 kg, and, sub-
sequently apply the relation of the cooling volume, which is 50 m^3
for the air conditioning and 800 m^3 for the cooling house, i.e.:
311 kg * 800 m^3 / 50 m^3 = 4966 kg of food content (apples, yay ^.^).

The larger cooling houses are determined similar, just assuming
different volumes and energetic values.

"""

from pathlib import Path
from typing import Any, cast

import numpy as np
import pandas as pd

from pysimmods.base.types import ModelInitVals, ModelParams

BASE_DATA = (Path(__file__) / ".." / "base_data.csv").resolve()

PARAM_KEYS = (
    ["eta_percent"]
    + ["length_m", "width_m", "height_m", "isolation_m"]
    + ["lambda_w_per_m_k", "t_min_deg_celsius", "t_max_deg_celsius"]
)
INIT_KEYS = ["c_j_per_kg_k", "theta_t_deg_celsius", "mass_kg"]


def hvac_preset(
    pn_max_kw: float,
    randomize_params: bool = False,
    randomize_inits: bool = False,
    workload: float = 33,
    seed: int | None = None,
    **kwargs,
) -> tuple[ModelParams, ModelInitVals]:
    rng = np.random.default_rng(seed)
    data = pd.read_csv(BASE_DATA)

    data = _compute_dynamic_values(data)

    e_kwh_day = pn_max_kw * 24
    e_kwh_day_workload = e_kwh_day * workload / 100
    e_kwh_year = e_kwh_day_workload * 365

    closest_row = data.iloc[
        (data["energy_kwh_a"] - e_kwh_year).abs().idxmin()
    ].to_dict()

    if randomize_params:
        workload = int(rng.integers(25, 75))
        e_kwh_day_workload = e_kwh_day * workload / 100
        e_kwh_year = e_kwh_day_workload * 365

        energetic_value = float(rng.integers(30, 250))
        volume_m3 = e_kwh_year / energetic_value
        length_m, width_m, height_m = _sample_dimensions(volume_m3, rng)

        params: ModelParams = {
            "p_max_kw": pn_max_kw,
            "length_m": float(length_m),
            "width_m": float(width_m),
            "height_m": float(height_m),
            "isolation_m": float(rng.uniform(0.15, 0.25)),
            "lambda_w_per_m_k": float(rng.uniform(0.1, 0.2)),
            "eta_percent": (1 - (energetic_value / 250)) * 100,
            "t_min_deg_celsius": float(rng.uniform(-40.0, -27.0)),
            "t_max_deg_celsius": float(rng.uniform(-20.0, 17.5)),
        }
    else:
        params: ModelParams = {k: closest_row[k] for k in PARAM_KEYS}
        params["p_max_kw"] = pn_max_kw

    if randomize_inits:
        c_j_per_kg_k = float(rng.uniform(2000, 4000))
        cooling_ref = 2390 * float(rng.uniform(5, 15))
        inits = {
            "c_j_per_kg_k": c_j_per_kg_k,
            "theta_t_deg_celsius": float(
                rng.uniform(
                    cast("float", params["t_min_deg_celsius"]),
                    cast("float", params["t_max_deg_celsius"]),
                )
            ),
            "mass_kg": cooling_ref
            / c_j_per_kg_k
            * cast("float", params["length_m"])
            * cast("float", params["width_m"])
            * cast("float", params["height_m"]),
            "cooling": bool(rng.choice([True, False])),
        }
    else:
        inits = {k: closest_row[k] for k in INIT_KEYS}
        inits["cooling"] = True

    if randomize_params or randomize_inits:
        check_configuration(params, inits)

    return params, inits


def _compute_dynamic_values(data: pd.DataFrame) -> pd.DataFrame:
    data["surface_m2"] = data.eval(
        "2*(length_m*width_m + length_m*height_m + width_m*height_m)"
    )
    data["volume_m3"] = data.eval("length_m * width_m * height_m")
    data["energy_kwh_a"] = data.eval("volume_m3 * energetic_value")
    cooling_ref = 2390 * 500 / 50  # from the 2 kW device
    data["mass_kg"] = cooling_ref / data["c_j_per_kg_k"] * data["volume_m3"]
    data.at[0, "mass_kg"] = 5.0  # The fridge has a different load
    return data


def _sample_dimensions(
    volume_m3: float, rng: np.random.Generator
) -> tuple[float, float, float]:
    # Sample height first, constraint by cube root upper bound
    h_max = volume_m3 ** (1 / 3)
    height_m = rng.uniform(h_max * 0.4, h_max * 0.95)

    # Given height, max width is (volume / height) ** 0.5
    w_max = (volume_m3 / height_m) ** 0.5
    width_m = rng.uniform(height_m, w_max)

    # Given height and width, length is determined
    length_m = volume_m3 / (width_m * height_m)

    return round(length_m, 2), round(width_m, 2), round(height_m, 2)


def check_configuration(params: dict[str, Any], inits: dict[str, Any]):
    alpha = (
        params["lambda_w_per_m_k"]
        * params["length_m"]
        * params["width_m"]
        * params["height_m"]
        / params["isolation_m"]
    )
    eta = params["eta_percent"] / 100
    p = params["p_max_kw"] * 1000
    C = inits["c_j_per_kg_k"]
    m = inits["mass_kg"]

    # calculate temperature change under worst case conditions
    t_diff = 40 - (-22)
    delta_t = ((alpha * t_diff) - eta * p) / (C * m)

    if delta_t > -0.0001:
        # current configuration is not powerful enough to cool
        # down the warehouse
        cool_factor = (alpha * t_diff - (-0.0001) * C * m) / (eta * p)
        params["cool_factor"] = round(cool_factor / 5, 2)



"""Presets for the battery model."""

from typing import Any

import numpy as np

from pysimmods.base.types import ModelInitVals, ModelParams


def battery_preset(
    pn_max_kw: float,
    cap_kwh: float | None = None,
    soc_percent: float = 50.0,
    soc_min_percent: float = 15.0,
    pn_to_cap_ratio: float = 5.0,
    randomize_params: bool = False,
    randomize_inits: bool = False,
    randomize_eta_pc: bool = False,
    seed: int | None = None,
    **kwargs: dict[str, Any],
) -> tuple[ModelParams, ModelInitVals]:
    rng = np.random.default_rng(seed)
    if randomize_params:
        pn_to_cap_ratio = rng.uniform(1.0, 10.0)
        p_charge_max_kw = pn_max_kw * rng.uniform(0.5, 1.0)
        p_discharge_max_kw = pn_max_kw * rng.uniform(0.5, 1.0)
        soc_min_percent = rng.uniform(0.0, 25.0)
    else:
        p_charge_max_kw = pn_max_kw
        p_discharge_max_kw = pn_max_kw

    eta_pc = [-2.109566, 0.403556, 97.110770]

    if randomize_eta_pc:
        eta_pc = [pc * rng.uniform(0.9, 1.1) for pc in eta_pc]

    if cap_kwh is None:
        cap_kwh = pn_max_kw * pn_to_cap_ratio

    params: ModelParams = {
        "cap_kwh": cap_kwh,
        "soc_min_percent": soc_min_percent,
        "p_charge_max_kw": p_charge_max_kw,
        "p_discharge_max_kw": p_discharge_max_kw,
        "eta_pc": eta_pc,
    }

    if randomize_inits:
        soc_percent = rng.uniform(soc_min_percent, 100.0)

    inits: ModelInitVals = {"soc_percent": soc_percent}

    return params, inits
